#ifndef persistence_h_Fahlaewahgaebaqu
#define persistence_h_Fahlaewahgaebaqu

#include <cpptest.h>
#include <cpptest-suite.h>

#include <libstick-0.1/persistence.h>

using namespace libstick;


class persistence_TestSuite: public Test::Suite {

    private:
        typedef simplicial_complex<3, uint32_t, double> scomplex;
        typedef persistence<3, uint32_t, double> pers;
        typedef pers::boundary_matrix bm;
        typedef pers::lowestones_matrix lom;
        typedef pers::transformation_matrix tm;

        bool setupcalled;
        scomplex ball, ring, torus;
        scomplex::simplex_order oball, oring, otorus;

    public:
        persistence_TestSuite() :
            setupcalled(false),
            oball(ball),
            oring(ring),
            otorus(torus)
            {
            TEST_ADD(persistence_TestSuite::test_matrix_reduction);
            TEST_ADD(persistence_TestSuite::test_betti_numbers);
            TEST_ADD(persistence_TestSuite::test_lowestones);
            TEST_ADD(persistence_TestSuite::test_diagram);
        }

    protected:
        virtual void setup() {
            if (setupcalled)
                return;
            setupcalled = true;

            //   7 ------------------------------------- 8
            //    \                                     /
            //     \    4 ------------------------- 5  /
            //      \    \                         /  /
            //       \    \      1 ------- 2      /  /
            //        \    \      \       /      /  /
            //         \    \      \     /      /  /
            //          \    \      \   /      /  /
            //           \    \      \ /      /  /
            //            \    \      3      /  /
            //             \    \           /  /
            //              \    \         /  /
            //               \    \       /  /
            //                \    \     /  /
            //                 \    \   /  /
            //                  \    \ /  /
            //                   \    6  /
            //                    \     /
            //                     \   /
            //                      \ /
            //                       9

            scomplex::simplex ssring[] = {
                // dimension, faces, value...
                /*  1 */ {0, {},           1},
                /*  2 */ {0, {},           2},
                /*  3 */ {0, {},           3},
                /*  4 */ {0, {},           4},
                /*  5 */ {0, {},           5},
                /*  6 */ {0, {},           6},
                /*  7 */ {1, {2, 3},       6.01},
                /*  8 */ {1, {3, 1},       6.02},
                /*  9 */ {1, {1, 2},       6.03},
                /* 10 */ {1, {4, 5},       6.04},
                /* 11 */ {1, {5, 6},       6.05},
                /* 12 */ {1, {6, 4},       6.06},
                /* 13 */ {1, {1, 4},       6.07},
                /* 14 */ {1, {1, 5},       6.08},
                /* 15 */ {2, {13, 14, 10}, 6.0801},
                /* 16 */ {1, {3, 6},       6.09},
                /* 17 */ {1, {2, 5},       6.10},
                /* 18 */ {1, {2, 6},       6.11},
                /* 19 */ {2, {9, 14, 17},  6.1101},
                /* 20 */ {2, {7, 16, 18},  6.1102},
                /* 21 */ {2, {11, 17, 18}, 6.1103},
                /* 22 */ {1, {1, 6},       6.12},
                /* 23 */ {2, {12, 13, 22}, 6.1201},
                /* 24 */ {2, {8, 16, 22},  6.1202}
            };
            const size_t cntssring = sizeof(ssring)/sizeof(scomplex::simplex);
            ring.add_simplices(ssring, cntssring);
            oring.reset();

            scomplex::simplex sstorus[] = {
                // dimension, faces, value...
                /* 25 */ {0, {},           7},
                /* 26 */ {0, {},           8},
                /* 27 */ {0, {},           9},
                /* 28 */ {1, {25, 26},     9.01},
                /* 29 */ {1, {26, 27},     9.02},
                /* 30 */ {1, {25, 27},     9.03},
                /* 31 */ {1, {25, 1},      9.04},
                /* 32 */ {1, {26, 1},      9.05},
                /* 33 */ {2, {28, 31, 32}, 9.0501},
                /* 34 */ {1, {27, 1},      9.06},
                /* 35 */ {2, {30, 31, 34}, 9.0601},
                /* 36 */ {1, {27, 3},      9.07},
                /* 37 */ {2, {36, 34, 8},  9.0701},
                /* 38 */ {1, {27, 2},      9.08},
                /* 39 */ {1, {26, 2},      9.09},
                /* 40 */ {2, {9, 32, 39},  9.0901},
                /* 41 */ {2, {29, 38, 39}, 9.0902},
                /* 42 */ {2, {7, 38, 36},  9.0903},
                /* 43 */ {1, {4, 25},      9.10},
                /* 44 */ {1, {5, 26},      9.11},
                /* 45 */ {1, {6, 27},      9.12},
                /* 46 */ {1, {4, 26},      9.13},
                /* 47 */ {1, {4, 27},      9.14},
                /* 48 */ {1, {5, 27},      9.15},
                /* 49 */ {2, {43, 47, 30}, 9.1501},
                /* 50 */ {2, {12, 45, 47}, 9.1502},
                /* 51 */ {2, {29, 44, 48}, 9.1503},
                /* 52 */ {2, {48, 11, 45}, 9.1504},
                /* 53 */ {2, {10, 46, 44}, 9.1505},
                /* 54 */ {2, {43, 46, 28}, 9.1506},
            };
            const size_t cntsstorus = sizeof(sstorus)/sizeof(scomplex::simplex);
            torus = ring;
            torus.add_simplices(sstorus, cntsstorus);
            otorus.reset();

            scomplex::simplex ssball[] = {
                // dimension, faces, value...
                {0, {},                1},
                {0, {},                2},
                {0, {},                3},
                {0, {},                4},
                {0, {},                5},
                {1, {1, 2},            6},
                {1, {2, 3},            7},
                {1, {3, 4},            8},
                {1, {4, 1},            9},
                {1, {1, 5},           10},
                {1, {2, 5},           11},
                {1, {3, 5},           12},
                {1, {4, 5},           13},
                {2, {6, 10, 11},      14},
                {2, {7, 11, 12},      15},
                {2, {8, 12, 13},      16},
                {2, {9, 13, 10},      17},
                {1, {1, 3},           18},
                {2, {18, 6, 7},       19},
                {2, {18, 8, 9},       20},
                {2, {18, 10, 12},     21},
                {3, {21, 14, 15, 19}, 22},
                {3, {21, 16, 17, 20}, 23},
            };
            const size_t cntssball = sizeof(ssball)/sizeof(scomplex::simplex);
            ball.add_simplices(ssball, cntssball);
            oball.reset();
        }

        virtual void tear_down() {
        }

        void test_matrix_reduction() {
            pers ringp(oring);
            ringp.compute_matrices();
            const bm &ringbm = ringp.get_boundary_matrix();
            const bm &ringrbm = ringp.get_reduced_boundary_matrix();
            const tm &ringtm = ringp.get_transformation_matrix();
            TEST_ASSERT(ringrbm == boolean_colrowmatrix<uint32_t>(ringbm) * ringtm);


            pers torusp(otorus);
            torusp.compute_matrices();
            const bm &torusbm = torusp.get_boundary_matrix();
            const bm &torusrbm = torusp.get_reduced_boundary_matrix();
            const tm &torustm = torusp.get_transformation_matrix();
            TEST_ASSERT(torusrbm == boolean_colrowmatrix<uint32_t>(torusbm) * torustm);

            const lom &toruslo = torusp.get_lowestones_matrix();
            uint32_t torusloe_coords[] = {1, 0, 8, 7, 13, 10, 11, 0, 0, 0, 0,
                0, 24, 0, 15, 0, 21, 19, 20, 0, 0, 0, 23, 0, 0, 31, 28, 29, 0,
                0, 42, 0, 33, 0, 35, 0, 37, 0, 41, 40, 0, 0, 0, 0, 52, 50, 53,
                49, 51, 0, 0, 0, 0, 0, 0};
            lom torusloe(torusloe_coords, torusloe_coords + sizeof(torusloe_coords)/sizeof(uint32_t));
            TEST_ASSERT(torusloe == toruslo);


            pers ballp(oball);
            ballp.compute_matrices();
            const bm &ballbm = ballp.get_boundary_matrix();
            const bm &ballrbm = ballp.get_reduced_boundary_matrix();
            const tm &balltm = ballp.get_transformation_matrix();
            TEST_ASSERT(ballrbm == boolean_colrowmatrix<uint32_t>(ballbm) * balltm);

            const lom &balllo = ballp.get_lowestones_matrix();
            uint32_t ballloe_coords[] = {1, 0, 6, 7, 8, 10, 0, 0, 0, 17, 0, 14,
                15, 16, 0, 0, 0, 0, 19, 0, 23, 22, 0, 0};
            lom ballloe(ballloe_coords, ballloe_coords + sizeof(ballloe_coords)/sizeof(uint32_t));
            TEST_ASSERT(ballloe == balllo);

            //std::cout << std::endl;
            //ballp.interpret_reduction(std::cout);
            //torusp.compute_diagrams();
            //torusp.interpret_persistence(std::cout);
            //std::cout << std::endl;
            //std::cout << std::endl;
        }

        void test_betti_numbers() {
            pers ringp(oring);
            ringp.compute_diagrams();

            TEST_ASSERT(ringp.get_persistence_diagram(-1).betti() == 0);
            TEST_ASSERT(ringp.get_persistence_diagram(0).betti() == 0);
            TEST_ASSERT(ringp.get_persistence_diagram(1).betti() == 1);
            TEST_ASSERT(ringp.get_persistence_diagram(2).betti() == 0);
            TEST_ASSERT(ringp.get_persistence_diagram(3).betti() == 0);
            for (unsigned d=0; d < 3; ++d)
                TEST_ASSERT(ringp.get_persistence_diagram(d).persistent_betti(oring.size()-1, oring.size()-1) ==
                        ringp.get_persistence_diagram(d).betti());


            pers torusp(otorus);
            torusp.compute_diagrams();

            TEST_ASSERT(torusp.get_persistence_diagram(-1).betti() == 0);
            TEST_ASSERT(torusp.get_persistence_diagram(0).betti() == 0);
            TEST_ASSERT(torusp.get_persistence_diagram(1).betti() == 2);
            TEST_ASSERT(torusp.get_persistence_diagram(2).betti() == 1);
            TEST_ASSERT(torusp.get_persistence_diagram(3).betti() == 0);
            for (unsigned d=0; d < 3; ++d)
                TEST_ASSERT(torusp.get_persistence_diagram(d).persistent_betti(otorus.size()-1, otorus.size()-1) ==
                        torusp.get_persistence_diagram(d).betti());

            // Torus was made from ring. Hence, persistent betti numbers must correspond to betti numbers of ring.
            for (unsigned d=0; d < 3; ++d)
                TEST_ASSERT(torusp.get_persistence_diagram(d).persistent_betti(oring.size()-1, oring.size()-1) ==
                        ringp.get_persistence_diagram(d).betti());


            pers ballp(oball);
            ballp.compute_diagrams();

            //std::cout << std::endl;
            //ballp.interpret_reduction(std::cout);
            //std::cout << std::endl;

            TEST_ASSERT(ballp.get_persistence_diagram(-1).betti() == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(0).betti() == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(1).betti() == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(2).betti() == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(3).betti() == 0);
            for (unsigned d=0; d < 3; ++d)
                TEST_ASSERT(ballp.get_persistence_diagram(d).persistent_betti(oball.size()-1, oball.size()-1) ==
                        ballp.get_persistence_diagram(d).betti());

            // Before we inserted the two tetrahedra and the interior triangle, we had a sphere.
            TEST_ASSERT(ballp.get_persistence_diagram(-1).persistent_betti(oball.size()-4, oball.size()-4) == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(0).persistent_betti(oball.size()-4, oball.size()-4) == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(1).persistent_betti(oball.size()-4, oball.size()-4) == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(2).persistent_betti(oball.size()-4, oball.size()-4) == 1);
            TEST_ASSERT(ballp.get_persistence_diagram(3).persistent_betti(oball.size()-4, oball.size()-4) == 0);

            // Before we inserted the two tetrahedra, we had a double-sphere.
            TEST_ASSERT(ballp.get_persistence_diagram(-1).persistent_betti(oball.size()-3, oball.size()-3) == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(0).persistent_betti(oball.size()-3, oball.size()-3) == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(1).persistent_betti(oball.size()-3, oball.size()-3) == 0);
            TEST_ASSERT(ballp.get_persistence_diagram(2).persistent_betti(oball.size()-3, oball.size()-3) == 2);
            TEST_ASSERT(ballp.get_persistence_diagram(3).persistent_betti(oball.size()-3, oball.size()-3) == 0);
        }

        void test_lowestones() {
            test_lowestones_impl(oring);
            test_lowestones_impl(otorus);
            test_lowestones_impl(oball);
        }

        void test_lowestones_impl(const scomplex::simplex_order& so) {
            pers p(so);
            p.compute_matrices();
            const lom &lowestones = p.get_lowestones_matrix();

            // Check that lowestones contains every column-index only once,
            // except the zeros.
            lom cpy = lowestones;
            sort(cpy.begin(), cpy.end());
            lom::iterator it = cpy.begin();
            // Skip the zeros
            while (it != cpy.end() && *it == 0)
                ++it;
            // From now on, no column index appears twice
            for (; it+1 != cpy.end(); ++it)
                TEST_ASSERT(*it < *(it+1));

            for (unsigned r=0; r < lowestones.size(); ++r) {
                // If (r,c) is a one then
                //   (i) a cycle dies with c --> row c is zero
                //  (ii) a cycle is born with r --> column r is zero
                //Hence
                //   (i) the column r is a zero-column
                //   (i) the row c is a zero-column

                const uint32_t c = lowestones[r];
                if (c > 0) {
                    TEST_ASSERT(p.get_reduced_boundary_matrix().get(r,c) == true);
                    TEST_ASSERT(p.get_reduced_boundary_matrix().get_column(r).size() == 0);
                    TEST_ASSERT(lowestones[c] == 0);
                }
            }
        }

        void test_diagram() {
            test_diagram_impl(oring);
            test_diagram_impl(otorus);
            test_diagram_impl(oball);
        }

        void test_diagram_impl(const scomplex::simplex_order& so) {
            pers p(so);
            p.compute_diagrams();

            // Even Jesus died after his birth. Homology classes are nothing better.
            for (int d=-1; d <= 3; ++d) {
                const pers::diagram &dia = p.get_persistence_diagram(d);
                for (unsigned i=0; i < dia.births.size(); ++i)
                    if (dia.births[i].death > 0)
                        assert(dia.births[i].birth < dia.births[i].death);
            }

        }
};

#endif
